/** 
 *   Camera Control for STAR I.
 *
 *   (c) Copyright 2005 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include "my_stdio.h"

#include "camera.h"
#include "fits_io.h"
#include "camera_tools.h"
#include "version.h"
#include "novas.h"


site_info geo_loc;

int set_location(void)
{
	struct image_2d im;
	im.ft = (struct fit_tag *) malloc(sizeof(struct fit_tag));
	im.ft->next = NULL;
	ct_read_header_file(&im, GLOBAL_CONFIG_FILE);
	
	geo_loc.latitude = im.latitude;
	geo_loc.longitude = im.longitude;
	geo_loc.height = im.altitude;
	geo_loc.temperature = 20.0; /* not used */
	geo_loc.pressure = 1020.0; /* not used */

	fits_tag_free(im.ft);

	return 0;
}


double mod24(double x)
{
	while (x < 0) x += 24.0;
	while (x > 24.0) x -= 24.0;
			 
	return x;
}     


double strdeg2double(char *str)
{
	double g = 0.0,
		m = 0.0,
		s = 0.0;
	double sig = 1.0;
	char *s1, *s2;

	s1 = str;
	g = strtod(s1, &s2);
	if(g < 0) sig = -1.0;
	if( s2 != NULL ) {
		s1 = s2+1;
		m = strtod(s1, &s2);
		if(s2 != NULL){
			s1 = s2+1;
			s = strtod(s1, &s2);
		}
	}

	return g + (sig * s / 60.0 + sig * m)/60.0;
}


int ct_read_save_file(int cam, char *filename)
{
	int len ;
	int *data;
	int status;
	char str_tmp[80];
	struct image_2d im;
	struct tm *t;
	double rar, decr;

	im.ft = (struct fit_tag *) malloc(sizeof(struct fit_tag));
	im.ft->next = NULL;

	if ( cc_act_rcv_setup(cam, &im.s) < 0) {
		PVERB(PINFO, " Error reading image setup.\n");
		return CT_IM_SETUP;
	}
	len = im.s.naxis2 * im.s.naxis1;
	if ( im.s.naxis2 > CCD_Y_LEN || im.s.naxis2 <= 0 ||  /* 384 */
	     im.s.naxis1 > CCD_X_LEN || im.s.naxis1 <= 0 ) { /* 576 */
		PVERB(PINFO, " invalid dimension %d x %d data image\n",
		      im.s.naxis2, im.s.naxis1);
		return CT_IM_DIM;
	}
	data = (int *) malloc(len * sizeof(int));
	PVERB(PINFO, " geting %d x %d data image: ", im.s.naxis2, im.s.naxis1);
	if ( cc_act_rcv_data(cam, data, &len) < 0) {
		PVERB(PINFO, " Error reading image.\n");
		free(data);
		return CT_IM_READ;
	}
	im.data = data;
	sprintf(str_tmp,
		"stari camera control v %s, (c) Federico A. Bareilles (License GPL)", VERSION);
	comment_push(im.ft, str_tmp);
	fits_tag_push(im.ft, "SOFT_VER", VERSION, "Software version number");
	im.ra[0] = im.dec[0] = 0;
	sprintf(im.filter, "-");
	ct_read_header_file(&im, GLOBAL_CONFIG_FILE);
	t = gmtime( (time_t *) &im.s.tv.tv_sec);
	sprintf(im.date_obs,"%04d-%02d-%02d", t->tm_year + 1900, 
		t->tm_mon + 1, t->tm_mday);
	sprintf(im.time_obs,"%02d:%02d:%02d", t->tm_hour, 
		t->tm_min, t->tm_sec);
	im.tjd = julian_date((short int)(t->tm_year + 1900), 
			     (short int)(t->tm_mon + 1), 
			     (short int)t->tm_mday, 
			     (double)t->tm_hour + 
			     (((double)t->tm_sec)/60.0 + 
			      ((double)t->tm_min))/60.0);
	sidereal_time (im.tjd, 0.0, 0.0, &im.gst);
	im.lst = mod24(im.gst + im.longitude / 15.0);
	im.ha = mod24(im.lst - strdeg2double(im.ra));
	
	equ2hor (im.tjd, 32.184 + 38.0, 0.0, 0.0,
		 &geo_loc,
		 strdeg2double(im.ra), 
		 strdeg2double(im.dec), 0,
		 &im.z, &im.azimuth, &rar, &decr);
	im.azimuth += 180.0;
	while(im.azimuth > 360.0) im.azimuth -= 360.0;
	im.airmas = airmas(im.z);
/*
  HA
  dist polar 90 +- declinacion
*/
	status = image_2d_save(filename, &im);
	free(data);
	fits_tag_free(im.ft);
	if (status != 0) {
		PVERB(PINFO, " Fail.\n");
		fits_report_error(stdout, status);
		return CT_IM_FITS_ERR;
	}

	PVERB(PINFO, "ok.\n");
	return CT_IM_OK;
}


#define LINE_LEN 160

static int parse_line(char *line, char **tag, char **data, char **comment)
{
	int i, j;
	int len;
	
	len = strlen(line);
	if( len > LINE_LEN) len = LINE_LEN;
	for( i=0; (i < len) && (line[i] != '=');i++);
	if (i < len && *line != '#' && line != 0 ) {
		line[i] = 0;
		*tag = stripwhite(line);
		i++;
		for( j=i; (j < len) && (line[j] != '#');j++);		
		if( j <len){
			line[j] = 0;
			j++;
			*comment=stripwhite(line+j);
		}else{
			*comment = NULL;
		}
		*data = stripwhite(line+i);

	}else{
		*tag = NULL;
		*data = NULL;
		*comment = NULL;
			
		return -1;
	}

	return 0;

}

int ct_read_header_file(struct image_2d *im, char *filename)
{
	FILE *fp;
	char line[LINE_LEN];
	char *sline, *tag, *data, *comment;

	fp = fopen(filename, "r");
	
	if( fp == NULL ){
                SYSLOG(LOG_ERR,
                       "ct_read_header_file: Error opening file: %s",
		       filename);
                return -1;
        }
	
	memset(line, 0, sizeof(line));
	while( fgets(line, sizeof(line), fp)){
		if(line[strlen(line)-1] == '\n')
			line[strlen(line)-1] = 0;
		sline = stripwhite(line);
		if (parse_line(sline, &tag, &data, &comment) == 0){
			if( memcmp(tag,"OBJECT", 6) == 0 ){
				sprintf(im->object,"%s", data);
			}else if( memcmp(tag,"OBSERVER", 8) == 0 ){
				sprintf(im->observer, "%s", data);
			}else if( memcmp(tag,"OBSERVAT", 8) == 0 ){
				sprintf(im->observat, "%s", data);
			}else if( memcmp(tag,"TELESCOP", 8) == 0 ){
				sprintf(im->telescope, "%s", data);
			}else if( memcmp(tag,"INSTRUME", 8) == 0 ){
				sprintf(im->instrument, "%s", data);
			}else if( memcmp(tag,"RA", 2) == 0 ){
				sprintf(im->ra, "%s", data);
			}else if( memcmp(tag,"DEC", 3) == 0 ){
				sprintf(im->dec, "%s", data);
			}else if( memcmp(tag,"FILTER", 6) == 0 ){
				sprintf(im->filter, "%s", data);
			}else if( memcmp(tag,"LATITUDE", 8) == 0 ){
				im->latitude = strdeg2double(data);
			}else if( memcmp(tag,"LONGITUD", 8) == 0 ){
				im->longitude = strdeg2double(data);
			}else if( memcmp(tag,"ALTITUDE", 8) == 0 ){
				im->altitude = strdeg2double(data);
			}else if( memcmp(tag,"COMMENT", 7) == 0 ){
				comment_push(im->ft, data);
			}else{
				if(comment != NULL)
					fits_tag_push(im->ft, tag, data, 
						      comment);
				else
					fits_tag_push(im->ft, tag, data, "");
			}
		}
		memset(line, 0, sizeof(line));
	}

/*     
EPOCA = J2000
*/
	fclose(fp);

	return 0;
}
