/** 
 *   Camera Control for STAR I.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include <fitsio.h>
#include "fits_io.h"

int printerror( int status)
{
	/*****************************************************/
	/* Print out cfitsio error messages and exit program */
	/*****************************************************/

	if (status) {
		fits_report_error(stderr, status); /* print error report */
		return -1;
	}
	
	return 0;
}

static char *str_im_type[] = {
	"Undefined", /* IMT_NONE  */
	"OBJECT",    /* IMT_IMAGE */
	"DARK",      /* IMT_DARK  */
	"BIAS",      /* IMT_BIAS  */
	"FOCUS",     /* IMT_FOCUS */
	"Undefined", /* IMT_SHADE */
	"Undefined", /* IMT_CLEAR */
	"FLATT",     /* IMT_FLATT */
	""
};


int comment_push(struct fit_tag *ft, char *text)
{
	int len;
	struct fit_tag *ftp;

	ftp = ft;
	while(ftp->next != NULL) ftp = ftp->next;
	len = strlen(text)+1;
	if( len > FLEN_COMMENT) len = FLEN_COMMENT;
	memcpy(ftp->str, text, len);
	ftp->tag[0] = 0;
	ftp->next  = (struct fit_tag *) malloc(sizeof(struct fit_tag));
	ftp->next->next = NULL;
	
	return 0;
}

int fits_tag_push(struct fit_tag *ft, char *tag, char *text, char *comment)
{
	int len;
	struct fit_tag *ftp;

	ftp = ft;
	while(ftp->next != NULL) ftp = ftp->next;
	len = strlen(text)+1;
	if (len > 0) {
		if( len > FLEN_VALUE) len = FLEN_VALUE;
		memcpy(ftp->str, text, len);
	}
	len = strlen(tag)+1;
	if (len > 0) {
		if( len > FLEN_KEYWORD) len = FLEN_KEYWORD;
		memcpy(ftp->tag, tag, len);
	}
	len = strlen(comment)+1;
	if (len > 0) {
		if( len > FLEN_KEYWORD) len = FLEN_KEYWORD;
		memcpy(ftp->comment, comment, len);
	}

	ftp->next  = (struct fit_tag *) malloc(sizeof(struct fit_tag));
	ftp->next->next = NULL;
	
	return 0;
}


int fits_tag_free(struct fit_tag *ft)
{
	struct fit_tag *ftp = ft;

	while(ft != NULL){
		ftp = ftp->next;
		free(ft);
		ft = ftp;
	}

	return 0;
}


int image_2d_save(char *filename, struct image_2d *im)
{
	struct fit_tag *ftp;
	char file[_POSIX_PATH_MAX];
	fitsfile *fptr;  /* pointer to the FITS file, defined in fitsio.h */
	int status;
	long fpixel, nelements;
	int bitpix   =  USHORT_IMG; /* 16-bit unsigned short pixel values */
	long naxis    =   2;        /* 2-dimensional image                */  
    	long naxes[2] = { im->s.naxis2, im->s.naxis1 };

	sprintf(file,"!%s", filename);	
	status = 0; /* initialize status before calling fitsio routines */
	if (fits_create_file(&fptr, file, &status))
		printerror( status );
	
	if ( fits_create_img(fptr,  bitpix, naxis, naxes, &status) )
		printerror( status );
	
	if ( fits_delete_key(fptr, "COMMENT", &status) )
		printerror( status );
	if ( fits_delete_key(fptr, "COMMENT", &status) )
		printerror( status );
	if ( fits_write_date(fptr, &status) )
		printerror( status );

	fpixel = 1;  /* first pixel to write      */
	nelements = naxes[0] * naxes[1]; /* number of pixels to write */

	/* write the array of unsigned integers to the FITS file */
	if ( fits_write_img(fptr, TUINT, fpixel, nelements, im->data, 
			    &status) )
		printerror( status );

	if ( fits_update_key(fptr, TSTRING, "IMAGETYP", 
			     str_im_type[im->s.im_type],
			     "Type of observation", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "OBJECT", 
			     im->object,
			     "Name of object", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "RA", im->ra, "Catalog RA", 
			     &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "DEC", im->dec, "Catalog DEC", 
			     &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "DATE-OBS", im->date_obs,
			     "UT Date of observation (YYYY-MM-DD)", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "TIME-OBS", im->time_obs,
			     "Universal time (UTC) at exposure start", 
			     &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "UT", im->time_obs,
			     "Universal time (UTC) at exposure start", 
			     &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "JD", &im->tjd,
			     "", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "EXPOSURE", &im->s.exposure,
			     "Total Exposure Time", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "EXPTIME", &im->s.exposure,
			     "Total Exposure Time", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TINT, "GAIN", &im->s.gain,
			     "System gain factor 1 or 4", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "FILTER", im->filter,
			     "Filter description", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TINT, "TEMPERAT", &im->s.temperature,
			     "CCD Temperature", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "OBSERVER", im->observer,
			     "Observer(s)", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "OBSERVAT", im->observat,
			     "", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "TELESCOP", im->telescope,
			     "", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TSTRING, "INSTRUME", im->instrument,
			     "Instrument used for observation", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "LATITUDE", &im->latitude,
			     "", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "LONGITUD", &im->longitude,
			     "", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "ALTITUDE", &im->altitude,
			     "", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "ST", &im->lst,
			     "Local mean siderial time", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "HA", &im->ha,
			     "Hour Angle", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "Z", &im->z,
			     "Topocentric zenith distance", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "AZIMUTH", &im->azimuth,
			     "Topocentric azimuth", &status) )
		printerror( status );
	if ( fits_update_key(fptr, TDOUBLE, "AIRMASS", &im->airmas,
			     "Air mass", &status) )
		printerror( status );
	for(ftp = im->ft; ftp->next != NULL; ftp = ftp->next) {
		if ( ftp->tag[0] != 0)
			if ( fits_update_key(fptr, TSTRING, 
					     ftp->tag, ftp->str, 
					     ftp->comment, &status) )
				printerror( status );
	}
	for(ftp = im->ft; ftp->next != NULL; ftp = ftp->next) {
		if ( ftp->tag[0] == 0)
			if ( fits_write_comment(fptr, ftp->str, &status) )
				printerror( status );
	}

	if ( fits_close_file(fptr, &status) ) /* close the file */
		printerror( status );
	
	return status;
}

